@@ -105,7 +105,7 @@ oracleSession
 	struct envEntry *envp;
 	struct srvEntry *srvp;
 	struct connEntry *connp;
-	char pid[30], *nlscopy;
+	char pid[30], *nlscopy = NULL;
 	ub4 is_connected;
 
 	/* it's easier to deal with empty strings */
@@ -754,6 +754,8 @@ struct oraTable
 	{
 		/* allocate an oraColumn struct for the column */
 		reply->cols[i-1] = (struct oraColumn *)oracleAlloc(sizeof(struct oraColumn));
+		reply->cols[i-1]->pgname = NULL;
+		reply->cols[i-1]->pgattnum = 0;
 		reply->cols[i-1]->pgtype = 0;
 		reply->cols[i-1]->pgtypmod = 0;
 		reply->cols[i-1]->used = 0;
@@ -1476,7 +1478,7 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 		dvoid *value = NULL;
 		sb4 value_len = 0;
 		ub2 value_type = SQLT_STR;
-		OCIDateTime *timest;
+		OCIDateTime **timest;
 		OCINumber *number;
 		char *num_format, *pos;
 
@@ -1518,13 +1520,16 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 							oraMessage);
 					}
 
-					(OCINumber *)value = number;
+					value = (dvoid *)number;
 					value_len = sizeof(OCINumber);
 					value_type = SQLT_VNU;
 					break;
 				case BIND_TIMESTAMP:
+					/* timest must exist until the OCIStmtExecute call */
+					timest = (OCIDateTime **)oracleAlloc(sizeof(OCIDateTime *));
+
 					/* allocate timestamp descriptor */
-					if (OCIDescriptorAlloc((const dvoid *)session->envhp, (dvoid **)&timest,
+					if (OCIDescriptorAlloc((const dvoid *)session->envhp, (dvoid **)timest,
 						OCI_DTYPE_TIMESTAMP_TZ, 0, NULL) != OCI_SUCCESS)
 					{
 						oracleCloseStatement(session, NULL);
@@ -1537,7 +1542,7 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 					if (checkerr(
 						OCIDateTimeFromText((dvoid *)session->userhp, session->errhp,
 							(const OraText *)param->value, strlen(param->value), (const OraText *)NULL, (ub1)0,
-							(const OraText *)NULL, (size_t)0, timest),
+							(const OraText *)NULL, (size_t)0, *timest),
 						(dvoid *)session->errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
 					{
 						oracleCloseStatement(session, NULL);
@@ -1548,9 +1553,9 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 					}
 
 					/* store in param->value to be able to free it later */
-					(OCIDateTime *)param->value = timest;
+					param->value = (char *)*timest;
 
-					(OCIDateTime **)value = &timest;
+					value = (dvoid *)timest;
 					value_len = sizeof(OCIDateTime **);
 					value_type = SQLT_TIMESTAMP_TZ;
 					break;
